import boto3
from urllib.parse import urlencode, quote
from botocore.auth import SigV4Auth
from botocore.awsrequest import AWSRequest
from botocore.credentials import Credentials
import requests
import urllib3
import logging
import json
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry
from ruxit.api.base_plugin import RemoteBasePlugin
from ruxit.api.exceptions import ConfigException

logger = logging.getLogger(__name__)

class AMPPluginRemote(RemoteBasePlugin):

    lastMetricTimestamps = {}

    def query(self, **kwargs):
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
        config = kwargs['config']
        aws_access_key_id = config["aws_access_key_id"].strip()
        aws_secret_access_key = config["aws_secret_access_key"].strip()
        region_name = config["region_name"].strip()
        workspace = config["workspace"].strip()
        api_endpoint = config["api_endpoint"].strip()
        if not api_endpoint.endswith("/"):
            api_endpoint += "/";
        if not api_endpoint.startswith("http"):
            api_endpoint = "https://" + api_endpoint
        api_token = config["api_token"].strip()
        metrics = config["metrics"].strip()
        if aws_access_key_id == "":
            raise ConfigException("Please enter an AWS access key")
        if aws_secret_access_key == "":
            raise ConfigException("Please enter an AWS secret access key")
        if region_name == "":
            raise ConfigException("Please enter a region name")
        if workspace == "":
            raise ConfigException("Please enter a workspace")
        if api_endpoint == "":
            raise ConfigException("Please enter a Dynatrace API endpoint")
        if api_token == "":
            raise ConfigException("Please enter a Dynatrace API token")
        if metrics == "":
            raise ConfigException("Please enter at least one metric")
        metricTimestamps = {}

        session = boto3.Session(
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            region_name=region_name
        )

        ampClient = session.client('amp', verify=False)
        workspace = ampClient.describe_workspace(workspaceId = workspace)

        session = requests.Session()
        session.mount(workspace["workspace"]["prometheusEndpoint"], HTTPAdapter(max_retries=Retry(total=3, backoff_factor=1, status_forcelist=[408, 429, 500, 502, 503, 504])))

        # Use metadata to decide between gauge and count
        request = self.getSignedRequest(workspace["workspace"]["prometheusEndpoint"] + "api/v1/metadata", region_name, aws_access_key_id, aws_secret_access_key, session)
        metadata = request.json()

        payload = []
        
        metricMetadata = []
        for metric in metrics.split("\n"):
            metricType = "gauge"
            metricName = None
            if "#" in metric:
                metricName = metric.split("#")[0].replace(" ", "_").lower()
                metric = metric.split("#")[1]
            elif metric in metadata["data"]:
                metricType = metadata["data"][metric][0]["type"]
                if metric.endswith("_bits"):
                    metricUnit = "Bit"
                elif metric.endswith("_bytes"):
                    metricUnit = "Byte"
                elif metric.endswith("_kilobytes"):
                    metricUnit = "KiloByte"
                elif metric.endswith("_megabytes"):
                    metricUnit = "MegaByte"
                elif metric.endswith("_milliseconds"):
                    metricUnit = "MilliSecond"
                elif metric.endswith("_minutes"):
                    metricUnit = "Minute"
                elif metric.endswith("_seconds"):
                    metricUnit = "Second"
                elif metric.endswith("_hours"):
                    metricUnit = "Hour"
                elif metric.endswith("_days"):
                    metricUnit = "Day"
                else:
                    metricUnit = "Unspecified"
                metricDescription = metadata["data"][metric][0]["help"]
            request = self.getSignedRequest(workspace["workspace"]["prometheusEndpoint"] + "api/v1/query?" + urlencode({"query": metric}, quote_via=quote), region_name, aws_access_key_id, aws_secret_access_key, session)
            results = request.json()
            if "error" in results:
                logger.warn(f'Error querying {metric}, {results["error"]}')
            elif "data" in results and "result" in results["data"]:
                for result in results["data"]["result"]:
                    if "metric" in result:
                        if metricName == None and "__name__" in result["metric"]:
                            metricName = result["metric"]["__name__"]
                            del result["metric"]['__name__']
                        if metricName != None:
                            dimensions = result["metric"]
                            if "value" in result and len(result["value"]) == 2:
                                timestamp = int(result["value"][0] * 1000)
                                metricTimestamps[metricName] = timestamp
                                if metric in metadata["data"]:
                                    metricMetadata.append({"name": metricName, "description": metricDescription, "unit": metricUnit, "count": metricType == "counter"})
                                else:
                                    metricMetadata.append({"name": metricName, "description": f"Query executed: {metric}", "unit": "Unspecified", "count": metricType == "counter"})
                                if self.lastMetricTimestamps.get(metricName, 0) != timestamp:
                                    self.addMintMetric(metricName, dimensions, result["value"][1], timestamp, payload, metricType == "counter")

        self.sendMINTMetrics(payload, api_endpoint, api_token)
        self.setMetricMetadata(metricMetadata, api_endpoint, api_token)
        self.lastMetricTimestamps = metricTimestamps

    def getSignedRequest(self, url, region_name, aws_access_key_id, aws_secret_access_key, session):
        credentials = Credentials(
            aws_access_key_id,
            aws_secret_access_key
        )

        sigv4 = SigV4Auth(credentials, 'aps', region_name)
        request = AWSRequest(method='GET', url=url)
        sigv4.add_auth(request)
        prepped = request.prepare()
        return session.get(prepped.url, headers=prepped.headers)
            
    def addMintMetric(self, metricName, dimensions, value, timestamp, payload, count):
        dimStr = ""
        for dimKey, dimValue in dimensions.items():
            if dimValue != "":
                if isinstance(dimValue, str):
                    sanitizedValue = dimValue.replace("\\n", "").replace("\\t", "").replace("\\r", "").replace("\\", "").strip()
                    dimStr += f",{dimKey}=\"{sanitizedValue}\""
                else:
                    dimStr += f",{dimKey}={dimValue}"
        if count:
            value = "count," + value
        payload.append(f"aws.prometheus.{metricName}{dimStr} {value} {timestamp}")
            
    def sendMINTMetrics(self, payload, url, token):
        data = []
        for line in payload:
            if len(data) == 1000:
                self.doPostRequest(data, url, token)
                data = []
            data.append(line)
        if len(data) > 0:
            self.doPostRequest(data, url, token)
                
    def doPostRequest(self, data, url, token):
        headers = {"Authorization": "Api-Token " + token, "Content-Type": "text/plain"}
        response = requests.post(f"{url}api/v2/metrics/ingest", verify=False, headers=headers, data='\n'.join(data))
        if response.status_code != 202:
            logger.warn("Metric Ingest API call failed. Input: " + '\n'.join(data) + " Output: " + response.text)
        else:
            logger.info("Metric Ingest API call successful Output: " + response.text)
            
    def setMetricMetadata(self, metrics, url, token):
        try:
            post_body = []
            for metric in metrics:
                entry = {
                    "scope": f'metric-aws.prometheus.{metric["name"]}' + (".count" if metric["count"] else ""),
                    "schemaId": "builtin:metric.metadata",
                    "value": {
                        "displayName": metric["name"],
                        "description": metric['description'],
                        "unit": metric['unit'],
                        "tags": []
                    }
                }
                post_body.append(entry)
            
            if post_body:
                post_body = json.dumps(post_body)
                headers = {"Authorization": "Api-Token " + token, "Content-Type": "application/json; charset=utf-8"}
                response = requests.post(f"{url}api/v2/settings/objects", data=post_body, verify=False, headers=headers)

                if response.status_code != 200:
                    logger.error(f"Not able to set metadata for metrics. Error response: {response.text}")
                else:
                    logger.info("Set metadata for metrics successfully")

        except requests.exceptions.HTTPError as e:
            logger.exception(f'Failed polling: {url}. {e}')
        except requests.exceptions.RequestException as e:
            logger.exception(f'Failed polling: {url}. {e}')